@@ -29,8 +29,10 @@ struct _Output
     gunichar fMabove;
     gunichar fMpost;
     gunichar fLengthMark;
+    gunichar fVirama; /* to handle virama in sinhala split matras */
     glong    fMatraIndex;
     gulong   fMatraTags;
+    gboolean fMatraWordStart;
     glong    fMPreOutIndex;
 
     MPreFixups *fMPreFixups;
@@ -49,7 +51,7 @@ static void initOutput(Output *output, const glong *originalOffsets, gunichar *o
     output->fOutIndex    = 0;
     output->fMatraTags   = 0;
 
-    output->fMpre = output->fMbelow = output->fMabove = output->fMpost = output->fLengthMark = 0;
+    output->fMpre = output->fMbelow = output->fMabove = output->fMpost = output->fLengthMark = output->fVirama = 0;
 
     output->fMPreOutIndex = -1;
     output->fMPreFixups = mpreFixups;
@@ -68,17 +70,23 @@ static void saveMatra(Output *output, gunichar matra, IndicOTCharClass matraClas
         output->fMpost = matra;
     } else if (IS_LENGTH_MARK(matraClass)) {
         output->fLengthMark = matra;
+    } else if (IS_VIRAMA(matraClass)) {
+        output->fVirama = matra;
     }
 }
 
-static void noteMatra(Output *output, const IndicOTClassTable *classTable, gunichar matra, guint32 matraIndex, gulong matraTags)
+static void initMatra(Output *output, guint32 matraIndex, gulong matraTags, gboolean wordStart)
 {
-    IndicOTCharClass matraClass = indic_ot_get_char_class(classTable, matra);
-
-    output->fMpre = output->fMbelow = output->fMabove = output->fMpost = output->fLengthMark = 0;
+    output->fMpre = output->fMbelow = output->fMabove = output->fMpost = output->fLengthMark = output->fVirama = 0;
     output->fMPreOutIndex = -1;
     output->fMatraIndex = matraIndex;
     output->fMatraTags = matraTags;
+    output->fMatraWordStart = wordStart;
+}
+
+static gboolean noteMatra(Output *output, const IndicOTClassTable *classTable, gunichar matra)
+{
+    IndicOTCharClass matraClass = indic_ot_get_char_class(classTable, matra);
 
     if (IS_MATRA(matraClass)) {
         if (IS_SPLIT_MATRA(matraClass)) {
@@ -94,7 +102,10 @@ static void noteMatra(Output *output, const IndicOTClassTable *classTable, gunic
         } else {
             saveMatra(output, matra, matraClass);
         }
-    }
+	
+	return TRUE;
+    } else
+      return FALSE;
 }
 
 static void noteBaseConsonant(Output *output)
@@ -117,8 +128,12 @@ static void writeChar(Output *output, gunichar ch, guint32 charIndex, gulong cha
 static void writeMpre(Output *output)
 {
     if (output->fMpre != 0) {
+        gulong tags = output->fMatraTags;
+	if (output->fMatraWordStart)
+	    tags &= ~init;
+	    
         output->fMPreOutIndex = output->fOutIndex;
-        writeChar(output, output->fMpre, output->fMatraIndex, output->fMatraTags);
+        writeChar(output, output->fMpre, output->fMatraIndex, tags);
     }
 }
 
@@ -150,6 +165,13 @@ static void writeLengthMark(Output *output)
     }
 }
 
+static void writeVirama(Output *output)
+{
+    if (output->fVirama != 0) {
+	writeChar(output, output->fVirama, output->fMatraIndex, output->fMatraTags);
+    }
+}
+
 static glong getOutputIndex(Output *output)
 {
     return output->fOutIndex;
@@ -163,6 +185,7 @@ glong indic_ot_reorder(const gunichar *chars, const glong *utf8_offsets, glong c
     MPreFixups *mpreFixups = NULL;
     Output output;
     glong i, prev = 0;
+    gboolean last_in_word = FALSE;
 
     if (outMPreFixups && (class_table->scriptFlags & SF_MPRE_FIXUP)) {
 	mpreFixups = indic_mprefixups_new (char_count);
@@ -184,10 +207,16 @@ glong indic_ot_reorder(const gunichar *chars, const glong *utf8_offsets, glong c
         }
 
         matra = vmabove - 1;
-        noteMatra(&output, class_table, chars[matra], /*matra*/ prev, blwf_p);
+	initMatra(&output, prev, blwf_p, !last_in_word);
+        while (noteMatra(&output, class_table, chars[matra]) &&
+               matra != prev)
+            matra--;
 
+	last_in_word = TRUE;
         switch (indic_ot_get_char_class(class_table, chars[prev]) & CF_CLASS_MASK) {
         case CC_RESERVED:
+  	    last_in_word = FALSE;
+	    /* Fall through */
         case CC_INDEPENDENT_VOWEL:
         case CC_ZERO_WIDTH_MARK:
             for (i = prev; i < syllable; i += 1) {
@@ -211,6 +240,7 @@ glong indic_ot_reorder(const gunichar *chars, const glong *utf8_offsets, glong c
             writeMabove(&output);
             writeMpost(&output);
             writeLengthMark(&output);
+	    writeVirama(&output);
             break;
 
         case CC_CONSONANT:
@@ -294,7 +324,10 @@ glong indic_ot_reorder(const gunichar *chars, const glong *utf8_offsets, glong c
 
 	    for (i = baseLimit; i < baseConsonant; i += 1) {
 		gunichar ch = chars[i];
-		gulong tag = blwf_p;
+		/* Applying blwf to the first consonant doesn't makes sense
+		 * since the below-form follows the consonant that it is
+		 * put under */
+		gulong tag = (i == baseLimit) ? half_p : blwf_p;
 		IndicOTCharClass charClass = indic_ot_get_char_class(class_table, ch);
 
 		if (IS_CONSONANT(charClass)) {
@@ -370,11 +403,10 @@ glong indic_ot_reorder(const gunichar *chars, const glong *utf8_offsets, glong c
 	    }
 
 	    /* write post-base consonants */
-	    /* FIXME: does this put the right tags on post-base consonants? */
 	    if (baseConsonant != lastConsonant) {
 		if (postBase <= lastConsonant) {
 		    for (i = postBase; i <= lastConsonant; i += 1) {
-			writeChar(&output, chars[i], /*i*/ prev, nukt_p);
+			writeChar(&output, chars[i], /*i*/ prev, pstf_p);
 		    }
 
 		    /* write halant that was after base consonant */
@@ -393,6 +425,7 @@ glong indic_ot_reorder(const gunichar *chars, const glong *utf8_offsets, glong c
 	    }
 
 	    writeLengthMark(&output);
+	    writeVirama(&output);
 
 	    /* write reph */
 	    if ((class_table->scriptFlags & SF_REPH_AFTER_BELOW) == 0) {
